@@ -14,21 +14,19 @@
    limitations under the License.
 """
 import gzip
-import tarfile
-from itertools import permutations
-from os import listdir, makedirs, path, rename
-from os.path import basename, getmtime, join
-from shutil import rmtree
+import shutil
+import sqlite3
+from os import listdir
+from os.path import getmtime, join, splitext
 from typing import List, Tuple
 
 import requests
 import torch
 from torch import Tensor
-from torch.nn.functional import dropout2d
 from tqdm import tqdm
 
-from neural_semigroups.constants import CURRENT_DEVICE, GAP_PACKAGES_URL
-from neural_semigroups.magma import Magma
+from neural_semigroups.constants import CURRENT_DEVICE
+from neural_semigroups.magma import Magma, get_two_indices_per_sample
 
 # the Cayley table of Klein Vierergruppe
 # pylint: disable=not-callable
@@ -65,53 +63,6 @@ def random_semigroup(
     return associative, magma.cayley_table
 
 
-def check_filename(filename: str) -> int:
-    """
-    checks filename, raises if it's incorrect
-
-    :param filename: filename to check
-    :returns: magma cardinality extracted from the filename
-    """
-    if isinstance(filename, str):
-        base_filename = basename(filename)
-        filename_parts = base_filename.split(".")
-        if len(filename_parts) == 3:
-            if (
-                filename_parts[0] in ("semigroup", "monoid", "group")
-                and filename_parts[2] == "zip"
-                and filename_parts[1].isdigit()
-            ):
-                return int(filename_parts[1])
-    raise ValueError(
-        "filename should be of format"
-        f"[semigroup|monoid|group].[int].zip, not {filename}"
-    )
-
-
-def check_smallsemi_filename(filename: str) -> int:
-    """
-    checks a filename from a `smallsemi` package, raises if it's incorrect
-
-    :param filename: filename from a `smallsemi` package to check
-    :returns: magma cardinality extracted from the filename
-    """
-    if isinstance(filename, str):
-        filename_parts = basename(filename).split(".")
-        if len(filename_parts) == 3:
-            if (
-                filename_parts[2] == "gz"
-                and filename_parts[1] == "gl"
-                and filename_parts[0][:-1] == "data"
-                and filename_parts[0][-1].isdigit()
-            ):
-                cardinality = int(filename_parts[0][-1])
-                if 2 <= cardinality <= 7:
-                    return cardinality
-    raise ValueError(
-        "filename should be of format data[2-7].gl.gz" f" not {filename}"
-    )
-
-
 def get_magma_by_index(cardinality: int, index: int) -> Magma:
     """
     find a magma from a lexicographical order by its index
@@ -138,12 +89,12 @@ def get_magma_by_index(cardinality: int, index: int) -> Magma:
     )
 
 
-def import_smallsemi_format(lines: List[bytes]) -> torch.Tensor:
+def import_smallsemi_format(lines: List[str]) -> torch.Tensor:
     """
     imports lines in a format used by ``smallsemi`` `GAP package`.
     Format description:
 
-    * filename is of a form ``data[n].gl.gz``, :math:`1<=n<=7`
+    * filename is of a form ``data[n].gl``, :math:`1<=n<=7`
     * lines are separated by a pair of symbols ``\\r\\n``
     * there are exactly :math:`n^2` lines in a file
     * the first line is a header starting with '#' symbol
@@ -160,7 +111,7 @@ def import_smallsemi_format(lines: List[bytes]) -> torch.Tensor:
 
     """
     raw_tables = torch.tensor(
-        [list(map(int, list(line.decode("utf-8")[:-1]))) for line in lines[1:]]
+        [list(map(int, list(line[:-1]))) for line in lines[1:]]
     ).transpose(0, 1)
     tables = torch.cat(
         [torch.zeros([raw_tables.shape[0], 1], dtype=torch.long), raw_tables],
@@ -170,57 +121,26 @@ def import_smallsemi_format(lines: List[bytes]) -> torch.Tensor:
     return tables.reshape(tables.shape[0], cardinality, cardinality)
 
 
-def get_equivalent_magmas(cayley_tables: torch.Tensor) -> torch.Tensor:
-    """
-    given a Cayley tables batch generate Cayley tables of isomorphic and
-    anti-isomorphic magmas
-
-    :param cayley_tables: a batch of Cayley tables
-    :returns: a batch of Cayley tables of isomorphic and anti-isomorphic magmas
-    """
-    equivalent_cayley_tables = list()
-    for permutation in permutations(range(cayley_tables.shape[1])):
-        permutation_tensor = torch.tensor(
-            permutation, device=cayley_tables.device
-        )
-        sample_index, one, two = get_two_indices_per_sample(
-            cayley_tables.shape[0], cayley_tables.shape[1]
-        )
-        isomorphic_cayley_tables = torch.zeros(
-            cayley_tables.shape, dtype=torch.long, device=cayley_tables.device
-        )
-        isomorphic_cayley_tables[
-            sample_index, permutation_tensor[one], permutation_tensor[two]
-        ] = permutation_tensor[cayley_tables[sample_index, one, two]]
-        equivalent_cayley_tables.append(isomorphic_cayley_tables)
-        anti_isomorphic_cayley_tables = torch.zeros(
-            cayley_tables.shape, dtype=torch.long, device=cayley_tables.device
-        )
-        anti_isomorphic_cayley_tables[
-            sample_index, permutation_tensor[one], permutation_tensor[two]
-        ] = permutation_tensor[cayley_tables[sample_index, two, one]]
-        equivalent_cayley_tables.append(anti_isomorphic_cayley_tables)
-    return torch.unique(torch.cat(equivalent_cayley_tables, dim=0), dim=0)
-
-
 def download_file_from_url(
     url: str, filename: str, buffer_size: int = 1024
 ) -> None:
     """
     downloads some file from the Web to a specified destination
 
-    >>> download_file_from_url("https://python.org/", "/tmp/test.html")
-    >>> import subprocess
-    >>> subprocess.run("ls /tmp/test.html", shell=True).returncode
-    0
+    >>> import os
+    >>> from neural_semigroups.constants import TEST_TEMP_DATA
+    >>> temp_file = os.path.join(TEST_TEMP_DATA, "test.html")
+    >>> if os.path.exists(temp_file):
+    ...     os.remove(temp_file)
+    >>> download_file_from_url("https://python.org/", temp_file)
+    >>> os.path.exists(temp_file)
+    True
 
     :param url: a valid HTTP URL
     :param filename: a valid filename
     :param buffer_size: a number of bytes to read from URL at once
     """
     response = requests.get(url, stream=True)
-    if not response.ok:
-        raise ValueError(f"Wrong response from URL: {response.status_code}")
     file_size = int(response.headers.get("Content-Length", 0))
     progress = tqdm(
         response.iter_content(chunk_size=buffer_size),
@@ -261,47 +181,21 @@ def find_substring_by_pattern(
     raise ValueError(f"pattern {starts_with}.*{ends_before} not found")
 
 
-def download_smallsemi_data(data_path: str) -> None:
-    """
-    downloads, unzips and moves ``smallsemi`` data
-
-    :param data_path: data storage path
-    :returns:
-    """
-    full_name_with_version = find_substring_by_pattern(
-        strings=requests.get(GAP_PACKAGES_URL).text.split("\n"),
-        starts_with="smallsemi",
-        ends_before=".tar.bz2",
-    )
-    temp_path = path.join(data_path, "tmp")
-    rmtree(temp_path, ignore_errors=True)
-    makedirs(temp_path, exist_ok=True)
-    archive_path = path.join(temp_path, f"{full_name_with_version}.tar.bz2")
-    download_file_from_url(
-        url=f"{GAP_PACKAGES_URL}{full_name_with_version}.tar.bz2",
-        filename=archive_path,
-    )
-    with tarfile.open(archive_path) as archive:
-        archive.extractall(temp_path)
-    rename(
-        path.join(temp_path, full_name_with_version, "data", "data2to7"),
-        path.join(data_path, "smallsemi_data"),
-    )
-
-
 def get_newest_file(dir_path: str) -> str:
     """
     get the last modified file from a diretory
 
     >>> from pathlib import Path
-    >>> rmtree("/tmp/tmp/", ignore_errors=True)
-    >>> makedirs("/tmp/tmp/")
-    >>> Path("/tmp/tmp/one").touch()
+    >>> from os import makedirs, path
+    >>> from neural_semigroups.constants import TEST_TEMP_DATA
+    >>> shutil.rmtree(path.join(TEST_TEMP_DATA, "tmp"), ignore_errors=True)
+    >>> makedirs(path.join(TEST_TEMP_DATA, "tmp"))
+    >>> Path(path.join(TEST_TEMP_DATA, "tmp", "one")).touch()
     >>> from time import sleep
     >>> sleep(0.01)
-    >>> Path("/tmp/tmp/two").touch()
-    >>> get_newest_file("/tmp/tmp/")
-    '/tmp/tmp/two'
+    >>> Path(path.join(TEST_TEMP_DATA, "tmp", "two")).touch()
+    >>> get_newest_file(path.join(TEST_TEMP_DATA, "tmp"))
+    'test_temp_data/tmp/two'
 
     :param dir_path: a directory path
     :returns: the last modified file's name
@@ -312,28 +206,6 @@ def get_newest_file(dir_path: str) -> str:
     )
 
 
-def get_two_indices_per_sample(
-    batch_size: int, cardinality: int
-) -> Tuple[Tensor, Tensor, Tensor]:
-    """
-    generates all possible combination of two indices
-    for each sample in a batch
-
-    >>> get_two_indices_per_sample(1, 2)
-    (tensor([0, 0, 0, 0]), tensor([0, 0, 1, 1]), tensor([0, 1, 0, 1]))
-
-    :param batch_size: number of samples in a batch
-    :param cardinality: number of possible values of an index
-    :returns: triples (index, index, index)
-    """
-    a_range = torch.arange(cardinality)
-    return (
-        torch.arange(batch_size).repeat_interleave(cardinality * cardinality),
-        a_range.repeat_interleave(cardinality).repeat(batch_size),
-        a_range.repeat(cardinality).repeat(batch_size),
-    )
-
-
 def make_discrete(cayley_cubes: torch.Tensor) -> torch.Tensor:
     """
     transforms a batch of probabilistic Cayley cubes and in the following way:
@@ -358,7 +230,7 @@ def make_discrete(cayley_cubes: torch.Tensor) -> torch.Tensor:
              [[1., 0.],
               [0., 1.]]]])
 
-    :param cayley_cubes: a batch of probabilistic cubes representing Cayley tables
+    :param cayley_cube: a batch of probabilistic cubes representing Cayley tables
     :returns: a batch of probabilistic cubes filled in with ``0`` or ``1``
     """
     cardinality = cayley_cubes.shape[1]
@@ -375,48 +247,6 @@ def make_discrete(cayley_cubes: torch.Tensor) -> torch.Tensor:
     return cube
 
 
-def load_data_and_labels_from_file(
-    database_filename: str,
-) -> Tuple[Tensor, Tensor]:
-    """
-    reads data from a special file format
-
-    :param database_filename: a special file to read data from
-    :returns: (a tensor with Cayley tables, a tensor of their labels)
-    """
-    check_filename(path.basename(database_filename))
-    torch_zip_file = torch.load(database_filename)
-    database = torch_zip_file["database"]
-    labels = torch_zip_file.get(
-        "labels", torch.zeros(len(database), dtype=torch.int64)
-    )
-    return database, labels
-
-
-def load_data_and_labels_from_smallsemi(
-    cardinality: int, data_path: str
-) -> Tuple[Tensor, Tensor]:
-    """
-    Loads data from ``smallsemi`` package
-
-    :param cardinality: which ``smallsemi`` file to use
-    :param data_path: where to search for ``smallsemi`` data
-    :returns: (a tensor with Cayley tables, a tensor of their labels)
-    """
-    filename = path.join(
-        data_path, "smallsemi_data", f"data{cardinality}.gl.gz"
-    )
-    check_smallsemi_filename(filename)
-    if not path.exists(filename):
-        download_smallsemi_data(data_path)
-    with gzip.open(filename, "rb") as file:
-        database = import_smallsemi_format(file.readlines()).to(CURRENT_DEVICE)
-    labels = torch.ones(
-        len(database), dtype=torch.int64, device=CURRENT_DEVICE
-    )
-    return database, labels
-
-
 def count_different(one: Tensor, two: Tensor) -> Tensor:
     """
     given two batches of the same size, counts number of positions in these
@@ -437,32 +267,33 @@ def count_different(one: Tensor, two: Tensor) -> Tensor:
     )
 
 
-def corrupt_input(cayley_cubes: Tensor, dropout_rate: float) -> Tensor:
+def hide_cells(cayley_table: Tensor, number_of_cells: int) -> Tensor:
     """
-    changes several cells in a Cayley table with uniformly distributed
-    random variables
+    set several cells in a Cayley table to math:`-1`
 
-    :param cayley_cubes: a batch of representations of Cayley tables
-                         as probability distributions
-    :param dropout_rate: a percentage of cells to distort
-    :returns: a batch of distorted Cayley cubes
+    >>> torch.manual_seed(42) # doctest: +ELLIPSIS
+    <torch...
+    >>> hide_cells(torch.tensor([[0, 1], [2, 3]]), 2).cpu()
+    tensor([[ 0,  1],
+            [-1, -1]])
+
+    :param cayley_table: a Cayley table
+    :param number_of_cells: a number of cells to hide
+    :returns: a Cayley table with hidden cells
     """
-    cardinality = cayley_cubes.shape[1]
-    return (
-        (1 - dropout_rate)
-        * dropout2d(
-            cayley_cubes.view(-1, cardinality * cardinality, cardinality, 1,)
-            - 1 / cardinality,
-            dropout_rate,
-            dropout_rate > 0,
-        )
-        + 1 / cardinality
-    ).view(-1, cardinality, cardinality, cardinality)
+    partial_table = cayley_table.clone()
+    cardinality = cayley_table.shape[0]
+    pairs = torch.randperm(cardinality * cardinality)[:number_of_cells]
+    partial_table[pairs // cardinality, pairs % cardinality] = -1
+    return partial_table
 
 
 def read_whole_file(filename: str) -> str:
     """
-    reads the whole file into a string
+    reads the whole file into a string,  for example
+
+    >>> read_whole_file("README.md").split("\\n")[2]
+    '# Neural Semigroups'
 
     :param filename: a name of the file to read
     :returns: whole contents of the file
@@ -475,7 +306,15 @@ def read_whole_file(filename: str) -> str:
 def partial_table_to_cube(table: Tensor) -> Tensor:
     """
     create a probabilistic cube from a partially filled Cayley table
-    ``-1`` is translated to :math:`\frac1n` where :math:`n` is table's cardinality
+    ``-1`` is translated to :math:`\\frac1n` where :math:`n` is table's
+    cardinality, for example
+
+    >>> partial_table_to_cube(torch.tensor([[0, -1], [0, 0]])).cpu()
+    tensor([[[1.0000, 0.0000],
+              [0.5000, 0.5000]],
+    <BLANKLINE>
+             [[1.0000, 0.0000],
+              [1.0000, 0.0000]]])
 
     :param table: a Cayley table, partially filled by ``-1``'s
     :returns: a probabilistic cube
@@ -490,4 +329,61 @@ def partial_table_to_cube(table: Tensor) -> Tensor:
     cube[rows, cols, table[rows, cols]] = 1.0
     rows, cols = torch.where(torch.eq(table, -1))
     cube[rows, cols, :] = 1 / cardinality
-    return cube.reshape([-1, cardinality, cardinality, cardinality])
+    return cube.reshape([cardinality, cardinality, cardinality])
+
+
+def create_table_if_not_exists(
+    database_name: str, table_name, columns: List[str]
+) -> None:
+    """
+    create a table if it does not exist
+
+    :param database_name: where to create a table
+    :param table_name: what table to create
+    :param columns: a list of strings of format "COLUMN_NAME COLUMN_TYPE"
+    :returns:
+    """
+    with sqlite3.connect(database_name, isolation_level=None) as connection:
+        connection.execute("PRAGMA journal_mode=WAL;")
+        cursor = connection.cursor()
+        cursor.execute(
+            "SELECT COUNT(*) FROM sqlite_master WHERE name = ?", [table_name]
+        )
+        if cursor.fetchone()[0] == 0:
+            connection.execute(
+                f"CREATE TABLE {table_name}({', '.join(columns)})"
+            )
+        cursor.close()
+
+
+def insert_values_into_table(
+    database_name: str, table_name: str, values: Tuple[str]
+) -> None:
+    """
+    inserts values into a table
+
+    :param database_name:
+    :param table_name:
+    :param values:
+    :returns:
+    """
+    with sqlite3.connect(database_name, isolation_level=None) as connection:
+        connection.execute("PRAGMA journal_mode=WAL;")
+        cursor = connection.cursor()
+        placeholders = ", ".join(len(values) * ["?"])
+        cursor.execute(
+            f"INSERT INTO {table_name} VALUES({placeholders})", values
+        )
+        cursor.close()
+
+
+def gunzip(archive_path: str) -> None:
+    """
+    extracts a GZIP file in the same folder
+
+    :param archive_path: a path ending with ``.gz``
+    :returns:
+    """
+    with gzip.open(archive_path, "rb") as f_in:
+        with open(splitext(archive_path)[0], "wb",) as f_out:
+            shutil.copyfileobj(f_in, f_out)
